package ssw.mj;

/**
 * A <code>Token</code> represents a terminal symbol.
 * Tokens are provided by the scanner for the parser.
 * They hold additional information about the symbol.
 *
 * @author AW
 */
public class Token {

	/*----- Token codes */
	public static final int
		none       =  0,
		ident      =  1,
		number     =  2,
		charConst  =  3,
		plus       =  4,
		minus      =  5,
		times      =  6,
		slash      =  7,
		rem        =  8,
		eql        =  9,
		neq        = 10,
		gtr        = 11,
		geq        = 12,
		less       = 13,
		leq        = 14,
		and        = 15,
		or         = 16,
		assign     = 17,
		pplus      = 18,
		mminus     = 19,
		semicolon  = 20,
		comma      = 21,
		period     = 22,
		lpar       = 23,
		rpar       = 24,
		lbrack     = 25,
		rbrack     = 26,
		lbrace     = 27,
		rbrace     = 28,
		break_     = 29,
		class_     = 30,
		do_        = 31,
		else_      = 32,
		final_     = 33,
		if_        = 34,
		new_       = 35,
		print      = 36,
		program    = 37,
		read       = 38,
		return_    = 39,
		void_      = 40,
		while_     = 41,
		eof        = 42;

	/** Token class (none, ident, ...). */
	int kind;

	/** Line number of this token. */
	int line;

	/** Column number of this token. */
	int col;

	/** Value of this token (for numbers or character constants). */
	int val;

	/** String representation of this token. */
	String str;

	public Token (int kind, int line, int col) {
		this.kind = kind; this.line = line; this.col = col;
	}

	public Token (int kind, int line, int col, int val, String string) {
		this(kind, line, col);
		this.val = val; this.str = string;
	}

	public boolean equals (Object obj) {
		if (this == obj) return true;  // same object
		if (!(obj instanceof Token)) return false;  // not a token
		return equals((Token) obj);
	}

	public boolean equals (Token other) {
		if (kind != other.kind || line != other.line || col != other.col)
			return false;

		switch (kind) {
			case ident:
				return ((str == null || str.length() == 0) &&
				        (other.str == null || other.str.length() == 0)) ||
				       (str != null && str.equals(other.str));
			case number:  // assert str is not null or empty
				return val == other.val && str.equals(other.str);
			case charConst:
				return val == other.val;
		}
		return true;
	}

	/**
	 * Returns a string representation of this Token object.
	 * This only includes the relevant attributes of the token.
	 */
	public String toString () {
		StringBuffer sb = new StringBuffer();
		sb.append("line " + line + ", col " + col + ": ");
		sb.append(Messages.getString(kind));
		switch (kind) {
			case ident: sb.append(" = "); sb.append(str); break;
			case number:
				sb.append(" = "); sb.append(val);
				sb.append(" (\""); sb.append(str); sb.append("\")");
				break;
			case charConst:
				sb.append(" = '"); sb.append((char) val); sb.append("'"); break;
		}
		return sb.toString();
	}
}
