package ssw.mj.codegen;

import ssw.mj.Messages;

/**
 * MircoJava Jump Label.
 * ====================
 * Forward jumps use labels which are still undefined.
 * Such jump instructions are linked in the code.
 * A fixup is made as soon as the label becomes defined.
 * Backward jumps use labels which are already defined.
 * Such jump instructions are already generated with the definite label address.
 * 
 * @author AW
 */
public class Label {
	/**	Is the jump destination already defined? */
	private boolean defined;

	/**	Jump destination address or head of unresolved forward jumps. */
	private int adr;
	
	/** Generates code for a jump to this label. */
	public void put () {
		if (defined) {
			// jump destination already known
			Code.put2(adr - (Code.pc-1));
		} else { 
			// insert code address into list of unresolved forward jumps
			Code.put2(adr); 
			adr = Code.pc-2; 
		}
	}
	
	/**	Defines <code>this</code> label to be at the current pc position */
	public void here () {
		if (defined) {  // should never happen
			throw new Error(Messages.getString("LAB_DEF"));
		}

		fixup(Code.pc);
	}
	
	/** Defines <code>this</code> label to be at the position 
	 *	of <code>dest</code>. */
	public void setTo (Label dest) {
		if (defined) {  // should never happen
			throw new Error(Messages.getString("LAB_DEF"));
		}
		if (!dest.defined) {  // should never happen
			throw new Error(Messages.getString("DEST_UNDEF"));
		}

		fixup(dest.adr);
	}

	// fixup of forward jumps to jump to destPC
	private void fixup (int destPC) {
		while (adr != 0) {
			int pos = adr;
			adr = Code.get2(adr);
			Code.put2(pos, destPC - (pos-1));
		}
		defined = true; adr = destPC;	
	}
}
